
package use_trafficintersection;

import java.awt.*;

/**
 * @author Sjoerd Westendorp
 */

/*
This class defines and draw a slider.
Listening to mouse events happens in RenderWorld

Java Swing does have a JSlider, but it ridiculously large.
Also it would then need it own listeners. Program flow would end up
all over the place.
*/

public class Slider {
    // Constants
    public static final int MARGIN = 5;
    public static final int MARKSIZE = 10;
    
    // Main variables
    public int x,y,w,h;
    public String title;
    public float min,max,start;
    public float state;
    public boolean isPrecentage;
    
    public Slider(int px, int py, int pw, int ph, String text, float fmin, float fmax, float fstart){
	x = px; y = py; w = pw; h = ph;
	title = text;
	min = fmin; max = fmax; start = fstart;
	
	isPrecentage = (min == 0f && max == 100f);
	
	state = start;
    }
    
    public void reset(){
	state = start;
    }
    
    public void draw(Graphics g){
	// Line location
	int liney = y+ 3*h/4;
	
	// Line
	g.setColor(Color.DARK_GRAY);
	g.drawLine(x+MARGIN, liney, x+w-MARGIN, liney);
	
	// Get State location
	float perc = (state-min)/(max-min);
	int loc = (int) ((w-2*MARGIN)*(perc));
	
	// Draw state
	g.setColor(Color.red);
	g.fillRect(x+MARGIN+loc-1, liney - MARKSIZE/2, 3, MARKSIZE);
	
	// Text
	String disptext = title + ": " + ((int)state);
	if(isPrecentage)
	    disptext += "%";
	
	g.setColor(Color.black);
	g.drawString(disptext,x+MARGIN,y+12);
    }
    
    public boolean mouseClick(int mx, int my){
	/*
	Check whther the click is on this slider.
	If not return false.
	If true return true and update state.
	*/
	// Check if we are on the slider
	if(mx < x || my < y || mx >= x+w || my >= y+h)
	    return false;
	// Check for edge cases
	if(mx <= x+MARGIN){
	    state = min;
	    return true;
	}
	if(mx >= x+w-MARGIN){
	    state = max;
	    return true;
	}
	// Main case
	float pos_x = mx-x-MARGIN;
	float prec = pos_x/(w-2*MARGIN);
	state = prec*(max-min) + min;
	
	return true;
    }
}
