
package use_trafficintersection.algorithm;

import java.awt.Color;
import java.util.ArrayList;
import use_trafficintersection.RenderWorld;
import use_trafficintersection.Renderer;

/**
 * @author Sjoerd Westendorp
 */

/*
This class defines the cars

Constructor: Widht, Heigth, starting direction, relative direction

Movement:
Algorithm controls "a".
"a" causes "v" causes "s"
Then "x" and "y" has to be calculated from "s".
Except if you have a better idea
*/

public class Car {
 
    boolean isAuto;
    boolean isDone = false;
    long loopSpawned, loopExpected;
    
    // Settings
    public static boolean usePlatooning = false;
    
    // Other object
    Lights lights;
    
    // Car in front
    private Car carInFront = null;
    private float carInFrontDis = 0;
    
    // Constants
    public static final int STATE_PRE_JUNCTION = 1;
    public static final int STATE_ON_JUNCTION = 2;
    public static final int STATE_POST_JUNCTION = 4;
    
    public static final float NEAR_TO_CAR = 3*Main.CAR_LENGHT;
    public static final float CLOSE_TO_CAR = 1.5f*Main.CAR_LENGHT;
    public static final float NEAR_TO_LIGHT = 5*Main.CAR_LENGHT;
    public static final float CLOSE_TO_LIGHT = 2*Main.CAR_LENGHT;
    
    public static final float V_MIN = 0;
    public static final float V_MAX = 1.0f;
    public static final float V_CRUISE = 0.95f * V_MAX;
    public static final float V_APPROACH = 0.5f;
    public static final float A_MIN = -V_MAX/30; // 0.5 second to zero
    public static final float A_MAX = V_MAX/180; // 3 seconds to max
    
    // Forward check
    public static final float FORWARDCHECK_STEP = 0.5f*Main.CAR_LENGHT;
    public static final float FORWARDCHECK_RANGE = Main.CAR_LENGHT;
    public static final int FORWARDCHECK_LISTSIZE = 20;
    public static final float FORWARDCHECK_MAX = FORWARDCHECK_LISTSIZE*FORWARDCHECK_STEP;
    
    private final float forwardCheckListX[] = new float[FORWARDCHECK_LISTSIZE];
    private final float forwardCheckListY[] = new float[FORWARDCHECK_LISTSIZE];
    
    // Corner constants
    public static final float CORNER_R = (float)(Renderer.JUNCSIZE - Renderer.LANE_R_OFFSET) * (float)Math.PI/2.0f;
    public static final float CORNER_R_MOD = 0.87f;
    public static final float CORNER_L = (float)(RenderWorld.LANESIZE + 2*Renderer.LANE_OFFSET) * (float)Math.PI/2.0f;
    public static final float CORNER_L_MOD = 1.03f;
    
    // Properties
    public int dirStart; // Direction it starts in (UP,DOWN,LEFT,RIGHT)
    public int dirEnd; // Direction it want to go in (UP,DOWN,LEFT,RIGHT)
    public int dirRel; // Relative direction it want to go in (GO_LEFT,GO_RIGHT,GO_FORWARD)
    public float rotation_start;
    public final int hitRadius;
    public final int hitRadius2;
    public int light; // Light the car has to wait on
    public int w,h;
    public Color c;
    
    // Movement
    public float x,y,s; // Coordinates
    public float rotation;
    public float a,v; // a and v are in the direction of "rotation"
    public int state = STATE_PRE_JUNCTION;
    
    // -----------------------------
    //		SETUP
    // -----------------------------
    
    public Car(Lights l, int tw, int th, int dir_start, int dir_rel, boolean auto, long loop){
	lights = l;
	w = tw; h = th;
	hitRadius = h/2;
	hitRadius2 = h;
	s = 0;
	setDirections(dir_start,dir_rel);
	setStartingPos(dir_start,dir_rel);
	if(auto)
	    c = Color.BLUE;
	else
	    c = Color.RED;
	isAuto = auto;
	loopSpawned = loop;
	// loopExpected
	float extraTimeNeeded = V_MAX/A_MAX;
	float extraTimeNeededDistance = 0.5f*A_MAX*extraTimeNeeded*extraTimeNeeded;
	float timeNeeded = V_MAX*(2.0f*(float)Renderer.START_OFFSET - extraTimeNeededDistance);
	loopExpected = loopSpawned + (long)(timeNeeded+extraTimeNeeded);
    }
    
    private void setDirections(int dir_start, int direction){
	dirStart = dir_start; 
	dirRel = direction; 
	/*
	It is posseble for dir end to be either
	- A direction (UP,DOWN,LEFT,RIGHT)
	- Or a relative direction (GO_LEFT,GO_RIGHT,GO_FORWARD)
	*/
	if( dirRel == Dir.GO_FORWARD){
	    if( dirStart == Dir.UP ){ dirEnd = Dir.DOWN; }
	    else if( dirStart == Dir.DOWN ){ dirEnd = Dir.UP; }
	    else if( dirStart == Dir.LEFT ){ dirEnd = Dir.RIGHT; }
	    else if( dirStart == Dir.RIGHT ){ dirEnd = Dir.LEFT; }
	    else{ dirEnd = Dir.ERROR; }
	}
	else if( dirRel == Dir.GO_LEFT){
	    if( dirStart == Dir.UP ){ dirEnd = Dir.RIGHT; }
	    else if( dirStart == Dir.DOWN ){ dirEnd = Dir.LEFT; }
	    else if( dirStart == Dir.LEFT ){ dirEnd = Dir.UP; }
	    else if( dirStart == Dir.RIGHT ){ dirEnd = Dir.DOWN; }
	    else{ dirEnd = Dir.ERROR; }
	}
	else if( dirRel == Dir.GO_RIGHT){
	    if( dirStart == Dir.UP ){ dirEnd = Dir.LEFT; }
	    else if( dirStart == Dir.DOWN ){ dirEnd = Dir.RIGHT; }
	    else if( dirStart == Dir.LEFT ){ dirEnd = Dir.DOWN; }
	    else if( dirStart == Dir.RIGHT ){ dirEnd = Dir.UP; }
	    else{ dirEnd = Dir.ERROR; }
	}
	else{
	    dirEnd = Dir.ERROR;
	}
    }
    
    private void setStartingPos(int dir_start, int dir_rel){
	if(dir_start == Dir.UP){
	    if(dir_rel != Dir.GO_RIGHT){
		x = -Renderer.LANE_OFFSET;
		y = -Renderer.START_OFFSET;
		light = Lights.LIGHT_U;
	    }
	    else{
		x = -Renderer.LANE_R_OFFSET;
		y = -Renderer.START_OFFSET;
		light = Lights.LIGHT_RU;
	    }
	    rotation_start = 0;
	}
	else if(dir_start == Dir.LEFT){
	    if(dir_rel != Dir.GO_RIGHT){
		x = -Renderer.START_OFFSET;
		y = Renderer.LANE_OFFSET;
		light = Lights.LIGHT_L;
	    }
	    else{
		x = -Renderer.START_OFFSET;
		y = Renderer.LANE_R_OFFSET;
		light = Lights.LIGHT_RL;
	    }
	    rotation_start = (float) (3*Math.PI/2);
	}
	else if(dir_start == Dir.DOWN){
	    if(dir_rel != Dir.GO_RIGHT){
		x = Renderer.LANE_OFFSET;
		y = Renderer.START_OFFSET;
		light = Lights.LIGHT_D;
	    }
	    else{
		x = Renderer.LANE_R_OFFSET;
		y = Renderer.START_OFFSET;
		light = Lights.LIGHT_RD;
	    }
	    rotation_start = (float) (2*Math.PI/2);
	}
	else if(dir_start == Dir.RIGHT){
	    if(dir_rel != Dir.GO_RIGHT){
		x = Renderer.START_OFFSET;
		y = -Renderer.LANE_OFFSET;
		light = Lights.LIGHT_R;
	    }
	    else{
		x = Renderer.START_OFFSET;
		y = -Renderer.LANE_R_OFFSET;
		light = Lights.LIGHT_RR;
	    }
	    rotation_start = (float) (Math.PI/2);
	}
	rotation = rotation_start;
    }
    
    @Override
    public String toString(){
	String str = "CAR ";
	str += "[" + Main.renderer.loopCount + "]";
	str += "("+x + "," + y + ")";
	str += " | a: " + a + ", v: " + v;
	return str;
    }
    
    // -----------------------------
    //		PHYSICS
    // -----------------------------
    
    public void accelerate(float acc){
	// limit acceleration
	if( acc > A_MAX )
	    acc = A_MAX;
	else if( acc < A_MIN )
	    acc = A_MIN;
	
	// Set a and v
	a = acc;
	v += a;
	
	// Limit velocity
	if( v > V_MAX )
	    v = V_MAX;
	else if( v < V_MIN )
	    v = V_MIN;
	
	// Cars can drift due to floating point precision errora
	// This prevents that during standstill
	if( Math.abs(v) < 0.01f )
	    return;
	
	// Set s
	s += v;
    }
    
    public float accelFromVelocity(float vel){
	// Limit velocity
	if( vel > V_MAX )
	    vel = V_MAX;
	else if( vel < V_MIN )
	    vel = V_MIN;
	
	if(vel > v){ // Accelerate
	    return A_MAX;
	}
	else{ //Deacellerate
	    return A_MIN;
	}
    }
    
    // -----------------------------
    //		STEP
    // -----------------------------
    
    /*
    * HOW TO PLATOON:
    * (CLOSE_TO_CAR/carInFrontDis)*carInFront.v
    */
    
    public void step(ArrayList<Car> cars, ArrayList<Ped> peds) throws Main.CollisionException{
	checkBounds();
	measureAccel();
	findCarInFront(cars);
	
	if(state == STATE_PRE_JUNCTION)
	    stepPre();
	else if(state == STATE_ON_JUNCTION)
	    stepOn();
	else
	    stepPost();
    }
    
    public void stepPre() throws Main.CollisionException{
	float accelCarFront = carInFrontHandler();
	float accelLight = lightHandler();
	
	// get actual acceleration
	accelerate(Math.min(accelCarFront,accelLight));
	
	// Move
	if( dirStart == Dir.UP )
	    y+=v;
	else if( dirStart == Dir.DOWN )
	    y-=v;
	else if( dirStart == Dir.LEFT )
	    x+=v;
	else
	    x-=v;
	
	if( Math.abs(x) > Renderer.JUNCSIZE
		|| Math.abs(y) > Renderer.JUNCSIZE)
	    return;
	state = STATE_ON_JUNCTION;
	s = 0;
        
    }
    
    public void stepOn() throws Main.CollisionException{
	accelerate(carInFrontHandler());
	
	if(dirRel == Dir.GO_FORWARD){
	    if( dirStart == Dir.UP )
		y+=v;
	    else if( dirStart == Dir.DOWN )
		y-=v;
	    else if( dirStart == Dir.LEFT )
		x+=v;
	    else
		x-=v;
	    
	    // Check whether we are still on the junction
	    if( Math.abs(x) <= Renderer.JUNCSIZE && Math.abs(y) <= Renderer.JUNCSIZE )
		return;
	}
	else if(dirRel == Dir.GO_RIGHT){
	    x += v*Math.cos(rotation+(Math.PI/2.0f));
	    y += v*Math.sin(rotation+(Math.PI/2.0f));
	    rotation += CORNER_R_MOD * v/CORNER_R;
	
	    if( Math.abs(rotation-rotation_start) < Math.PI/2 )
		return;
	}
	else{ // dirRel == Dir.GO_LEFT
	    x += v*Math.cos(rotation+(Math.PI/2.0f));
	    y += v*Math.sin(rotation+(Math.PI/2.0f));
	    rotation -= CORNER_L_MOD * v/CORNER_L;
	
	    if( Math.abs(rotation-rotation_start) < Math.PI/2 )
		return;
	}
	
	state = STATE_POST_JUNCTION;
    }
    
    public void stepPost() throws Main.CollisionException{
	accelerate(carInFrontHandler());
	
	
	if( dirEnd == Dir.UP )
	    y-=v;
	else if( dirEnd == Dir.DOWN )
	    y+=v;
	else if( dirEnd == Dir.LEFT )
	    x-=v;
	else
	    x+=v;
    }
    
    // -----------------------------
    //		HANDLERS
    // -----------------------------
    
    public float carInFrontHandler() throws Main.CollisionException{
	float accelCarFront;
	if(carInFront != null){
	    if(carInFrontDis < CLOSE_TO_CAR){
		if(usePlatooning && carInFront.isAuto && isAuto && carInFront.v > v) // PLATOONING
		    accelCarFront = accelFromVelocity((CLOSE_TO_CAR/carInFrontDis)*carInFront.v);
		else
		    accelCarFront = accelFromVelocity(0);
	    }
	    else if(carInFrontDis < NEAR_TO_CAR){
		accelCarFront = accelFromVelocity(V_APPROACH);
	    }
	    else
		accelCarFront = accelFromVelocity(V_CRUISE);
	    
	    if(carInFrontDis < Main.CAR_LENGHT){
		throw Renderer.mainobject.new CollisionException(toString());
	    }
	}
	else{ 
	    accelCarFront = accelFromVelocity(V_CRUISE);
	}
	return accelCarFront;
    }
    
    public boolean greenWaveHasHappened = false;
    public boolean useGreenWaveSpeed = false;
    public float greenWaveSpeed = V_MAX;
    public void tellNextGreenWave(int time){
	float t = (float) time;
	greenWaveSpeed = (getDistanceToLight()-RenderWorld.SIDESIZE/2)/t;
	useGreenWaveSpeed = true;
	// Limiting
	if(greenWaveSpeed < V_MIN)
	    greenWaveSpeed = V_MIN;
	else if(greenWaveSpeed > V_MAX)
	    greenWaveSpeed = V_MAX;
    }
    
    private float getGreenWaveSpeed(){
	return greenWaveSpeed;
    }
    
    public float lightHandler(){
	float accelLight;
	float distanceToLight = getDistanceToLight();
	// If auto, check for green wave
	if(isAuto && useGreenWaveSpeed){
	    // Check whether we missed it
	    if(greenWaveHasHappened && lights.getState(light) == Lights.RED)
		useGreenWaveSpeed = false;
	    // Check whether it is green right now
	    if(lights.getState(light) == Lights.GREEN){
		greenWaveHasHappened = true;
		if(distanceToLight < NEAR_TO_LIGHT)
		    return accelFromVelocity(V_MAX);
	    }
	    // Just continue on
	    return accelFromVelocity(getGreenWaveSpeed());
	}
	// Else do it the old way
	if(distanceToLight < NEAR_TO_LIGHT && lights.getState(light) == Lights.RED){
	    if(distanceToLight < CLOSE_TO_LIGHT ){
		accelLight = accelFromVelocity(0);
		if(Math.abs(v) < 0.01f)
		    lights.setLaneBool(light);
	    }
	    else{
		accelLight = accelFromVelocity(V_APPROACH);
	    }
	}
	else if(distanceToLight < NEAR_TO_LIGHT && lights.getState(light) == Lights.YELLOW){
	    accelLight = accelFromVelocity(V_APPROACH/2);
	} 
	else if(distanceToLight < CLOSE_TO_LIGHT && lights.getState(light) == Lights.YELLOW){
	    accelLight = accelFromVelocity(V_MAX);
	} 
	else{
	    accelLight = accelFromVelocity(V_MAX);
	}
	// return
	return accelLight;
    }
    
    // -----------------------------
    //		AVOIDANCE
    // -----------------------------
    
    private void findCarInFront(ArrayList<Car> cars){
	carInFrontDis = 0;
	carInFront = null;
	
	Car car;
	boolean couldBeOne;
	float currentBestDis = 99999999.0f, newDis;
	
	setForwardCheck();
	
	for(int i = 0; i < cars.size(); i++){
	    car = cars.get(i);
	    couldBeOne = false;
	    
	    // Lets not detect ourselves
	    if(car == this)
		continue;
	    
	    // Check whether we are in the same lane
	   //  or are going to be in the same lane
	    if(car.dirEnd == this.dirEnd)
		couldBeOne = true;
	    if(car.dirStart == this.dirStart)
		couldBeOne = true;
	    if((car.dirRel == Dir.GO_RIGHT) == (this.dirRel == Dir.GO_RIGHT))
		couldBeOne = true;
	    
	    if(!couldBeOne)
		continue;
	    
	    // Is it in front of us?
	    if(!forwardCheck(car))
		continue;
	    
	    // Now get our distance
	    newDis = Main.dis(car.x,car.y,this.x,this.y);
	    
	    // Do we have a winner?
	    if( newDis > currentBestDis)
		continue;
	    
	    // we have a winner!
	    currentBestDis = newDis;
	    carInFront = car;
	    carInFrontDis = newDis;
	}
    }
    
    private void setForwardCheck(){
	float nx = x, ny = y;
	nx += FORWARDCHECK_STEP*Math.cos(rotation+(Math.PI/2.0f));
	ny += FORWARDCHECK_STEP*Math.sin(rotation+(Math.PI/2.0f));
	for(int i = 0; i < FORWARDCHECK_LISTSIZE; i++){
	    nx += FORWARDCHECK_STEP*Math.cos(rotation+(Math.PI/2.0f));
	    ny += FORWARDCHECK_STEP*Math.sin(rotation+(Math.PI/2.0f));
	    
	    forwardCheckListX[i] = nx;
	    forwardCheckListY[i] = ny;
	}
    }
    
    private boolean forwardCheck(Car c){
	float nx, ny;
	float distanceSq;
	for(int i = 0; i < FORWARDCHECK_LISTSIZE; i++){
	    nx = forwardCheckListX[i];
	    ny = forwardCheckListY[i];
	    distanceSq = Main.disSq(c.x,c.y,nx,ny);
	    
	    if(distanceSq < FORWARDCHECK_RANGE*FORWARDCHECK_RANGE)
		return true;
	}
	return false;
    }
    
    public void drawForwardCheck(Renderer r){
	for(int i = 0; i < FORWARDCHECK_LISTSIZE; i++){
	    r.drawCross((int) forwardCheckListX[i],(int) forwardCheckListY[i]);
	}
    }
    
    // -----------------------------
    //		MEASURE
    // -----------------------------
    
    private void checkBounds(){
	if( dirEnd == Dir.UP && y < -Renderer.START_OFFSET )
	    isDone = true;
	if( dirEnd == Dir.DOWN && y > Renderer.START_OFFSET )
	    isDone = true;    
	if( dirEnd == Dir.LEFT && x < -Renderer.START_OFFSET )
	    isDone = true;
	if( dirEnd == Dir.RIGHT && x > Renderer.START_OFFSET )
	    isDone = true;
    }
    
    public float accel = 0;
    private float m_x0 = 9999999, m_y0 = 9999999;
    private float m_x1 = 9999999, m_y1 = 9999999;
    private float m_x2 = 9999999, m_y2 = 9999999;
    public void measureAccel(){
	
	// Move all coordinates one over
	m_x2 = m_x1;
	m_x1 = m_x0;
	m_x0 = x;
	m_y2 = m_y1;
	m_y1 = m_y0;
	m_y0 = y;
	
	if(m_x2 > 2*Renderer.WORLD_SIZE)
	    return; // not our time yet
	
	// Calc velocities
	float v01 = Main.dis(m_x0,m_y0,m_x1,m_y1);
	float v12 = Main.dis(m_x1,m_y1,m_x2,m_y2);
	
	// Calc acceleration
	float a012 = (float) Math.abs(v01-v12);
	accel += a012;
    }
    
    
    public float getDistanceToLight(){
	if(dirStart == Dir.UP){
	    return -Renderer.JUNCSIZE-y;
	}
	else if(dirStart == Dir.LEFT){
	    return -Renderer.JUNCSIZE-x;
	}
	else if(dirStart == Dir.DOWN){
	    return -Renderer.JUNCSIZE+y;
	}
	else if(dirStart == Dir.RIGHT){
	    return -Renderer.JUNCSIZE+x;
	}
	return 1270127.0f;
    }
}
